#ifdef OSR5
#pragma comment(exestr, "%Z% %M% %I% %E% %Q%")
#endif

#ifdef SOLARIS
#pragma ident	"%Z% %M% %I% %D% SMI"
#endif
/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<errno.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/dirent.h>
#if defined(SOLARIS) || defined(OSR5)
#include	<sys/stropts.h>
#endif
#include	<sys/file.h>
#include	<sys/uio.h>
#include	<sys/time.h>
#include	<sys/select.h>
#include	<sys/stat.h>
#include	<sys/stropts.h>

#include	"lxrun.h"
#include	"linux/fcntl.h"
#include	"devmap.h"
#include	"trace.h"
#ifdef OLD_PROCFS
#include	"procfs.h"
#endif

#ifdef OLD_WAY	/* ## */
#ifdef UW7
#include <fileio.h>
#if 0
#define	IOCPARM_MASK	0xff		/* parameters must be < 256 bytes */
#define	IOC_OUT		0x40000000	/* copy out parameters */
#define	IOC_IN		0x80000000	/* copy in parameters */
#define	_IOR(x,y,t)	(IOC_OUT|((((int)sizeof(t))&IOCPARM_MASK)<<16)|(x<<8)|y)
#define	_IOW(x,y,t)	(IOC_IN|((((int)sizeof(t))&IOCPARM_MASK)<<16)|(x<<8)|y)

#define	SIOCSPGRP	_IOW('s',  8, int)	/* set process group */
#define	SIOCGPGRP	_IOR('s',  9, int)	/* get process group */

#define	FIOSETOWN	_IOW('f', 124, int)	/* set owner */
#define	FIOGETOWN	_IOR('f', 123, int)	/* get owner */
#endif
#endif
#endif  /* OLD_WAY */

static int lx2os_oflags(int lx_flags);
static int os2lx_oflags(int os_flags);

static const char *iobufstr(void *buf, int len);

/*
 * file i/o system calls
 */

/*
 *	open
 *	fcntl
 *	flock		(NOT IMPLEMENTED)
 *	__llseek
 *	readdir
 *	fdatasync	(PARTLY IMPLEMENTED)
 */

typedef struct
{
	int	lx_flag;
	int	os_flag;
} oflags_t;

static const oflags_t	oflags[] =
{
	{ LX_O_CREAT,	O_CREAT		},
	{ LX_O_EXCL,	O_EXCL		},
	{ LX_O_NOCTTY,	O_NOCTTY	},
	{ LX_O_TRUNC,	O_TRUNC		},
	{ LX_O_APPEND,	O_APPEND	},
	{ LX_O_NONBLOCK,O_NONBLOCK	},
	{ LX_O_SYNC,	O_SYNC		},
	{ LX_O_DIRECTORY, 0		},	/* don't care for now */
#if defined(SOLARIS) || defined(UW7)
	{ LX_FASYNC,	FASYNC		},
#endif
};

static int
lx2os_oflags(int lx_flags)
{
	int	os_flags;
    int		n;

	switch (lx_flags & LX_O_ACCMOD)
	{
		case LX_O_RDONLY:
			os_flags = O_RDONLY;
			break;

		case LX_O_WRONLY:
			os_flags = O_WRONLY;
			break;

		case LX_O_RDWR:
			os_flags = O_RDWR;
			break;

		default:
		    return enoany( EINVAL,
				    "lx2os_oflags - unknown LX_O_ACCMOD");
	}

	lx_flags &= ~LX_O_ACCMOD;

	for (n = 0; lx_flags && (n < sizeof(oflags)/sizeof(*oflags)); ++n)
	if (lx_flags & oflags[n].lx_flag)
	{
	    lx_flags &= ~oflags[n].lx_flag;
	    os_flags |= oflags[n].os_flag;
	}

	if (lx_flags)
	{
	    TRACE_MSG( (TRACE_OPEN | TRACE_FCNTL),
		       ("TRACE: lx2os_oflags: Unknown bits %#x", lx_flags));
    
	    return enoany(EINVAL, "lx2os_oflags - invalid flags");
	}

	return os_flags;
}

static int
os2lx_oflags(int os_flags)
{
	int	lx_flags;
	int	n;

	switch (os_flags & O_ACCMODE)
	{
		case O_RDONLY:
			lx_flags = LX_O_RDONLY;
			break;

		case O_WRONLY:
			lx_flags = LX_O_WRONLY;
			break;

		case O_RDWR:
			lx_flags = LX_O_RDWR;
			break;

		default:
		    return enoany( EINVAL, "lx2os_oflags - unknown O_ACCMOD");
	}

	os_flags &= ~O_ACCMODE;

	for (n = 0; os_flags && (n < sizeof(oflags)/sizeof(*oflags)); ++n)
	if (os_flags & oflags[n].os_flag)
	{
	    os_flags &= ~oflags[n].os_flag;
	    lx_flags |= oflags[n].lx_flag;
	}

	if (os_flags)
	{
	    TRACE_MSG( (TRACE_OPEN | TRACE_FCNTL),
		       ("TRACE: os2lx_oflags: Unknown bits %#x", os_flags));
	    
	    return enoany(EINVAL, "lx2os_oflags");
	}

    return lx_flags;
}

static int
copy_in_flock(struct lx_flock *from, struct flock *to)
{
	if (! verify_read(from, sizeof(*from)))
		return -1;

	switch (from->l_type)
	{
		case LX_F_RDLCK:
			to->l_type = F_RDLCK;
			break;
		case LX_F_WRLCK:
			to->l_type = F_WRLCK;
			break;
		case LX_F_UNLCK:
			to->l_type = F_UNLCK;
			break;
		default:
		    return enoany(EINVAL,
				  "copy_in_flock - unknown from->l_type");
	}

	to->l_whence	= from->l_whence;
	to->l_start	= from->l_start;
	to->l_len	= from->l_len;
	to->l_pid	= from->l_pid;

	return 0;
}

static int
copy_out_flock(struct flock *from, struct lx_flock *to)
{
	if (! verify_write(to, sizeof(*to)))
		return -1;

	switch (from->l_type)
	{
		case F_RDLCK:
			to->l_type = LX_F_RDLCK;
			break;
		case F_WRLCK:
			to->l_type = LX_F_WRLCK;
			break;
		case F_UNLCK:
			to->l_type = LX_F_UNLCK;
			break;
		default:
		    return enoany(EINVAL,
				  "copy_out_flock - unknown from->l_type");
	}

	to->l_whence	= from->l_whence;
	to->l_start	= from->l_start;
	to->l_len	= from->l_len;
	to->l_pid	= from->l_pid;

	return 0;
}

#if defined(SOLARIS) || defined(OSR5)
/*
 * The Berkeley tty's don't have ptem pushed automatically.
 */
static int
prepare_slave(int fd)
{
	while (ioctl(fd, I_POP, 0) == 0);

	ioctl(fd, I_PUSH, "ptem");
	ioctl(fd, I_PUSH, "ldterm");
	ioctl(fd, I_PUSH, "ttcompat");

	return (fd);
}
#endif

int
lx_open(char *lx_path, int lx_flags, int mode)
{
    int 	os_flags;
    int 	fd;

    os_flags = lx2os_oflags(lx_flags);

    TRACE_MSG(TRACE_OPEN, ("TRACE: open(%s, %d -> %d, %#o)\n",
	       lx_path, lx_flags, os_flags, mode));

    if (os_flags == -1)
		return -1;

    if (((fd = open(lx_path, os_flags, mode)) != -1)
#ifdef OLD_PROCFS
	|| ((fd = lxprocfs_open(lx_path, os_flags, mode)) != -1)
#endif
	)
	{
#if defined(SOLARIS) || defined(OSR5)
	    if ( strlen(lx_path) >= sizeof ("/dev/tty") &&
		 strncmp(lx_path, "/dev/tty", sizeof ("/dev/tty") - 1) == 0)
		fd = prepare_slave(fd);
#endif
	    return fd;
	}
    

    return -1;
}

/* 
 * This was originally added to reconcile X paths, specifically
 * the NLS locale paths, and the app-default paths.
 * This handled now with lx_pathname mapping external to emulation fn but
 * keep for non phantom file procfs access() call.
 */

int
lx_access(char *lx_path, int mode)
{
    int	  	flags;
    int 	r;
  
    TRACE_MSG(TRACE_ACCESS, ("TRACE: access(%s, %#o)\n",  lx_path, mode));

    if (((r = access(lx_path, mode)) != -1)
#ifdef OLD_PROCFS
	|| ((r = lxprocfs_access(lx_path, mode)) != -1)
#endif
	)
	    return r;

    return -1;
}

int
lx_close(int fd)
{
#ifdef OLD_PROCFS
    if (!lxprocfs_close(fd))
	return 0;
#endif

    /*
     * Invalidate this file descriptor in any caches (devmap, mmap)
     */

    devmap_invalidate(fd);
    dz_invalidate(fd);
	
    return close(fd);
}

int
lx_read(int fd, void *buf, int len)
{
    int 		n;
    fd_set		readfds;
    struct timeval	tv;

#ifdef OLD_PROCFS
    if ((n = lxprocfs_read(fd, buf, len)) != -1)
	return n;
#endif

    READ_TRACE_MSG(("TRACE: read(%d, %s, %d) ", fd, iobufstr(buf, n), len));

    n = read(fd, buf, len);

    READ_TRACE_MSG(("--> %d\n", n));

    /*
     * This is to compensate for the strangeness with EOF on pipes.
     */

    if (n != 0)
	return n;

    FD_ZERO(&readfds);
    FD_SET(fd, &readfds);

    tv.tv_sec = 0;
    tv.tv_usec = 0;

    n = select(fd+1, &readfds, NULL, NULL, &tv);

    READ_TRACE_MSG(("TRACE: select for empty read returned %d\n", n));

    if (n != 1)
    {
	return enoany(EAGAIN,"TRACE: read select() failed");
    }

    READ_TRACE_MSG(("TRACE: retrying read\n"));
    return read(fd, buf, len);
}

int
lx_write(int fd, void *buf, int len)
{
	int	n;

#ifdef OLD_PROCFS
	if ((n = lxprocfs_write(fd, buf, len)) != -1)
	    return n;
#endif

	WRITE_TRACE_MSG(("TRACE: write(%d, %s, %d) ",
			 fd, iobufstr(buf, len), len));

	n = write(fd, buf, len);
	WRITE_TRACE_MSG(("--> %d\n", n));
	/*
	 * kludge for TCP/IP
	 */
	if ((n == -1) && (errno == EWOULDBLOCK) && (len > 4096))
	{
		errno = 0;
		n = write(fd, buf, 4096);
	}

	return n;
}

int
lx_readv(int fd, struct iovec *iov, int iovcnt)
{
    int 	n;

#ifdef OLD_PROCFS
    if ((n = lxprocfs_readv(fd, iov, iovcnt)) != -1)
	return n;
#endif

    n = readv(fd, iov, iovcnt);

    TRACE_CALL( TRACE_READ,
    {
	const struct iovec	*v;
	int			cnt;

	tprintf("TRACE: readv(fd %d", fd);

	for (v = iov, cnt = iovcnt; cnt; --cnt, ++v)
	    tprintf(", %s, len %d",
			iobufstr(v->iov_base, v->iov_len), v->iov_len);

	tprintf(") returned %d\n", n);
    });
    

    return n;
}

int
lx_writev(int fd, const struct iovec *iov, int iovcnt)
{
    int 	n;

#ifdef OLD_PROCFS
    if ((n = lxprocfs_writev(fd, iov, iovcnt)) != -1)
	return n;
#endif

    TRACE_CALL( TRACE_WRITE,
    {

	const struct iovec	*v;

	tprintf("TRACE: writev(fd %d", fd);

	for (v = iov, n = iovcnt; n; --n, ++v)
	    tprintf(", %s, len %d",
			    iobufstr(v->iov_base, v->iov_len), v->iov_len);

	tprintf(")\n");
    });
    


    if (iovcnt == 1)
    {
	int	st;
	char	*buf;

	WRITE_TRACE_MSG(("TRACE: using write for single writev\n"));

	st = write(fd, iov->iov_base, iov->iov_len);
	if ((st != -1) || (errno != ERANGE))
	    return st;

	WRITE_TRACE_MSG(("TRACE: Trying half step\n"));

	n = iov->iov_len;
	buf = (char *)(iov->iov_base);

	while (n)
	{
	    int 	len;

	    if (n > (8 * 1024))
		len = 8 * 1024;
	    else
		len = n;

	    if ((st = write(fd, buf, len)) != len)
		return st;

	    n -= len;
	    buf += len;
	}

	return iov->iov_len;
    }

    return writev(fd, iov, iovcnt);
}

int
lx_fcntl(int fd, int lx_cmd, int arg)
{
	int		r;
	int		flags;
	struct	flock	f;

#ifdef OLD_PROCFS
	if ((r = lxprocfs_fcntl(fd, lx_cmd, arg)) != -1)
	    return r;
#endif

	switch (lx_cmd)
	{
		case LX_F_DUPFD:
			return fcntl(fd, F_DUPFD, arg);

		case LX_F_GETFD:
			return fcntl(fd, F_GETFD, arg);

		case LX_F_SETFD:
			return fcntl(fd, F_SETFD, arg);

		case LX_F_GETFL:
		    if ((r = fcntl(fd, F_GETFL, arg)) == -1)
			return -1;

			flags = os2lx_oflags(r);
			FCNTL_TRACE_MSG(("TRACE: fcntl(%d, F_GETFL) %#x -> %#x\n",
							    fd, arg, flags));
			return flags;

		case LX_F_SETFL:
#if 0	/* Is this correct? */
			if (!verify_read((void *)arg, sizeof(int)))
			    flags = -1;
			else
#endif
			    flags = lx2os_oflags(arg & ~LX_O_ACCMOD);

			FCNTL_TRACE_MSG((
			    "TRACE: fcntl(%d, F_SETFL, %#x -> %#x)\n",
							    fd, arg, flags));
			if (flags == -1)
				return -1;
			return fcntl(fd, F_SETFL, flags);

		case LX_F_GETLK:
			if (copy_in_flock((struct lx_flock *)arg, &f) != 0)
				return -1;
			if ((r = fcntl(fd, F_GETLK, &f)) == -1)
				return -1;
			if (copy_out_flock(&f, (struct lx_flock *)arg) != 0)
				return -1;
			return r;

		case LX_F_SETLK:
			if (copy_in_flock((struct lx_flock *)arg, &f) != 0)
				return -1;
			return fcntl(fd, F_SETLK, &f);

		case LX_F_SETLKW:
			if (copy_in_flock((struct lx_flock *)arg, &f) != 0)
				return -1;
			return fcntl(fd, F_SETLKW, &f);

		case LX_F_GETOWN:
#ifdef OSR5
#ifdef OLD_WAY	/* ## */
		    FCNTL_TRACE_MSG(("TRACE: fcntl("
		                     "LX_F_GETOWN -> SIOCGPGRP\n"));

		    if (((r = fcntl(fd, SIOCGPGRP, NULL)) == -1) &&
			(errno == EINVAL))
		    {
			FCNTL_TRACE_MSG((
		      "TRACE: Retrying with fcntl(LX_F_GETOWN -> FIOGETOWN\n"));
			r = fcntl(fd, FIOGETOWN, NULL);
		    }

		    return r;
#endif
		    FCNTL_TRACE_MSG(("TRACE: fcntl(LX_F_GETOWN)"
		         "-> getpid()\n"));
		    return getpid();
#else
		    return fcntl(fd, F_GETOWN, arg);
#endif	/*  */		    

		case LX_F_SETOWN:
#ifdef OSR5
#ifdef OLD_WAY	/* ## */
		    FCNTL_TRACE_MSG(("TRACE: fcntl("
		                     "LX_F_SETOWN -> SIOCSPGRP,%d\n", arg);

		    if (((r = fcntl(fd, SIOCSPGRP, &arg)) == -1) &&
			(errno == EINVAL))
		    {
			FCNTL_TRACE_MSG((
		            "TRACE: Retrying with fcntl("
		            "LX_F_SETOWN -> FIOSETOWN, %d\n", arg));
			r = fcntl(fd, FIOSETOWN, &arg);
		    }

		    return r;
#endif
		    FCNTL_TRACE_MSG(("TRACE: fcntl(LX_F_SETOWN)"
		        " -> I_SETSIG\n"));
		    return ioctl(fd, I_SETSIG,
				 S_RDNORM|S_WRNORM|S_RDBAND|S_BANDURG);
#else
		    return fcntl(fd, F_SETOWN, arg);
#endif	/* OSR5 */
		    
		    
		default:
		    {
			char str[32];
			sprintf(str, "Unknown fcntl(0x%x) fd=%d", lx_cmd, fd);
			return enoany(ENOSYS, str);
		    }
    }
}

int lx_flock()		{ return enosys("flock"); }

int
lx__llseek(int fd, long off_hi, long off_lo, long *res, int whence)
{
	long	l;

	if (((off_lo >= 0) && (off_hi != 0)) ||
	    ((off_lo < 0) && (off_hi != -1)))
	{
	    return enoany( EINVAL, "lx_llseek - invalid off_{hi,lo}");
	}

	if (((l = lseek(fd, off_lo, whence)) == -1) ||
	!verify_write(res, 2 * sizeof(res[0])))
		return -1;

	res[0] = l;
	res[1] = (l < 0) ? -1 : 0;
	return 0;
}       

struct lx_dirent
{
	long	d_ino;
	long	d_off;
	short	d_namlen;
	char	d_name[1];
};

#define LX_D_NAME_SIZE	256	/* d_name[NAME_MAX+1] */

int
lx_readdir(int fd, struct lx_dirent *lx_dir, int count)
{
	char		buf[512];
	struct dirent	*d	= (struct dirent *)buf;

	if (getdents(fd, d, 512) <= 0)
		return -1;

	lx_dir->d_ino	= d->d_ino;
	lx_dir->d_off	= d->d_off;
	lx_dir->d_namlen= strlen(d->d_name);
	strncpy(lx_dir->d_name, d->d_name, LX_D_NAME_SIZE);
	lseek(fd, d->d_off, 0);

	return d->d_reclen;
}       


/*
** The behavior of fdatasync and fsync are slightly different, so
** this isn't a perfect implementation.
** The latter changes the access time in the inode.
** The former updates the file _without_ touching the inode.
** The main advantage of fdatasync over fsync is performance.
** Hopefully, this is the _only_ difference as far as most Linux
** programs are concerned.
*/
int
lx_fdatasync(int fd)
{
	return fsync(fd);
}


static const char *
iobufstr(void *buf, int len)
{
    static char		dbgbuf[80];
    char		*tp;
    u_char		*bp = (u_char *)buf;

    tp = dbgbuf;
    *tp++ = '"';
    while ((len > 0) && (tp < &dbgbuf[sizeof(dbgbuf)-10]))
    {
	u_char		c;

	--len;

	c = *bp++;
	if ((c >= ' ') && (c <= '~'))
	    *tp++ = c;
	else
	{
	    *tp++ = '\\';
	    *tp++ = '0' + ((c >> 6) & 0x03);
	    *tp++ = '0' + ((c >> 3) & 0x07);
	    *tp++ = '0' + (c & 0x07);
	}
    }

    if (len > 0)
    {
	*tp++ = '.';
	*tp++ = '.';
	*tp++ = '.';
    }

    *tp++ = '"';
    *tp = '\0';
    return dbgbuf;
}

