#ifdef OSR5
#pragma comment(exestr, "%Z% %M% %I% %E% %Q%")
#endif

#ifdef SOLARIS
#pragma ident	"%Z% %M% %I% %D% %Q% SMI"
#endif
/*
 * Copyright (C) 1996-1997 Michael Davidson.
 * All rights reserved.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<errno.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	<string.h>

#include	"lxrun.h"
#include	"version.h"
#include	"trace.h"

extern void lxrun_init(void);
extern int lx_execve(char *prog, char **argv, char **envp);

extern char 	**environ;
extern char	*linux_root;
extern char	*linux_uname;
extern char     emulator[];
extern int	quiet;

const char	options[] =
		    "t:uR:U:0:q";

static void setup_emul(const char *av0, const char *pathenv, char **emulp);
static const char *baseName(const char *file);


static void
usage(const char *prog)
{
    fprintf(stderr,
	"Linux Emulator\n"
	"Version %s, built on %s at %s\n"
	"\n"
	"Usage:\n"
	"    %s -u\n"
	"    %s "
		"[-t <element1>[,<element2>...]] "
		"[-R <root>]\n"
		"                [-U <uname>] [-q] "
		"prog [args ...]\n",
	    build_version, build_date, build_time,
	    prog, prog);

    trace_help();
    exit(1);
}


/*
 * LINUX emulator for x86 UNIX
 */
int
main(int argc, char **argv, char **envp)
{
	int	c;
	char	*pathenv = getenv("PATH");
	char    *emul;			/* Path to the lxrun binary. */
	char    *prog;			/* Path to the Linux program binary. */
	char    *argv0 = NULL;		/* argv[0] to call program with. */
	char	real_interp[PATH_MAX];
	int	ac = 0;
	char	**av;

	setup_emul(argv[0], pathenv, &emul);

	if (strcmp("lxrun", baseName(argv[0])) == 0)
	    while ((c = getopt(argc, argv, options)) != -1)
		switch (c)
		{
		    case 'q':		/* suppress warnings */
			quiet=1;
			break;

		    case 'u':		/* print usage message */
			usage(emul);
			break;

		    case 't':		/* enable syscall tracing */
			TRACE_INIT(optarg);
			break;

		    case 'U':		/* set value for uname to return */
			linux_uname = optarg;
			break;

		    case 'R':		/* set Linux root for path remapping */
			linux_root = optarg;
			break;

		    case '0':		/* explicitly specified argv0 to use */
			argv0 = optarg;
			break;

		default:
			usage(emul);
		}

	argc -= optind;
	argv += optind;
        if (argc < 1)    /* no filename arg */
		usage(emul);
		
	/*
	 * argv[0] is now Linux binary name or path.  But if we were
	 * called with a different argv0 (e.g. by lx_execve), stick that
	 * into the argv[] array now.
	 */
	prog = argv[0];
	if (argv0 != NULL)
		argv[0] = argv0;

	lxrun_init();

	chkpath_lxprog(pathenv, &prog);

	/*
	 * We use environ instead of envp in case we've already changed
	 * something in our environment that we want to pass onto our
	 * children.
	 */
	run_program(prog, argv, environ);

	/*
	 * like exec() a successful call to run_program() will transfer
	 * control to the new program and never return - if we get
	 * here then run_program() has failed in some way.  Try exec'ing
	 * the binary as if it were native.  We use lx_execve to handle
	 * shell script cases as well.
	 */

	lx_execve(prog, argv, environ);
	fatal(errno, "program load %s failed", prog);
	return errno;
}

static const char *
baseName(const char *file)
{
    int		n;

    for (n = 0; file[n]; ++n)
	    ;
    while((file[n] != '/') && (n >= 0))
	    --n;
    return &file[n+1];
}


/*
 * Init emulator[] with abs pathname of invoking binary if in PATH
 * emulator is preinitted with default config place
 */
static void
setup_emul(const char *av0, const char *pathenv, char **emulp)
{
    char    emulbf[PATH_MAX];		/* tmp storage for emulator name */
    char    *emul;			/* Path to the lxrun binary. */

    emulbf[0] = '\0';
    emul      = (char *)av0;   		/* invoking emulator binary */
    if(access(emul, X_OK) == 0)
    {
	strcpy(emulbf, emul);
    }
    else if (pathenv != NULL)	/* Check for lxrun binary in PATH */
    {
	/* We use a copy of pathenv since strtok is destructive. */
	char *pathcopy = strdup(pathenv);
	char *pathp    = pathcopy;

	/* Second set of ()s here stop gcc complaining */
	while ((pathp=strtok(pathp, ":"))) 
	{
	    strcpy(emulbf, pathp);
	    strcat(emulbf, "/");
	    strcat(emulbf, emul);

	    if(access(emulbf, X_OK) == 0)
		break;

	    emulbf[0] = '\0';	/* Couldn't find lxrun binary! */
	    pathp=NULL;
	}
	free(pathcopy);
    }
		
    if(emulbf[0] == '\0' || access(emulbf, X_OK) != 0)
    {
	fprintf(stderr, "%s: Can't find full path of lxrun binary!", emul);
	exit(1);
    }

    /* We want a full pathname to the emulator in case the Linux app
    ** changes directory and then does an exec call. */
    if(emulbf[0] != '/')
	{
	char *cwd = getcwd(NULL, PATH_MAX);
	strcpy(emulator, cwd);
	free(cwd);
	strcat(emulator, "/");
	}
    else
	emulator[0]='\0';
    strcat(emulator, emulbf);
    *emulp = emul;
}

