#ifdef OSR5
#pragma comment(exestr, "%Z% %M% %I% %E% %Q%")
#endif

#ifdef SOLARIS
#pragma ident	"%Z% %M% %I% %D% %Q% SMI"
#endif
/*
 * Copyright (C) 1997 Michael Davidson.
 * All rights reserved.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<errno.h>
#include	<string.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<ctype.h>
#include	<sys/stat.h>
#include	<unistd.h>

#include	"trace.h"
#include	"lxrun.h"

extern char	*linux_root;

#ifdef PATHMAPFILE
#include <malloc.h>

#ifdef PHANTOM_MAP_FILE
static int create_phantom_file(pathmap_t *pm, const char *from,
                                char *fn, int len)
#endif

typedef enum
{
    emt_UNDEF,
    NOMAP,
    PREFIX,
    DOMAP,
    DOEXEC
} map_t;

typedef struct pathmap_s pathmap_t;
struct pathmap_s
{
    map_t	type;       /* type of match */
    char	*from;      /* string to map from */
    size_t	fromlen;
    char	*to;        /* string to map to */
    pathmap_t	*next;      /* next in list */
};

pathmap_t	*pathmap_head = NULL;
#endif  /* PATHMAPFILE */

static const char *map_type_str(map_t type);

/*
 * safe getstring() for fetching pathnames from the
 * program being run under emulation - checks that
 * each page is readable before copying
 * returns total number of bytes copied (including the
 * null terminator) or -1 in the case of a fault
 */
static int
getstring(const void *src, void *dst, int len)
{
	register const char *s = (const char *)src;
	register       char *d = (char *)dst;

	while (len > 0)
	{
		register int n;

		if (getbyte(s) < 0)	/* check for faults */
			return -1;

		n    = PAGE_SIZE - ((unsigned long)src & (PAGE_SIZE-1));
		len -= n;

		while (--n >= 0)
			if (! (*d++ = *s++))
				return d - (char *)dst;
	}
	
	return d - (char *)dst;
}

/*
 * map Linux path names into system path names
 * (mainly for the benefit of the dynamic linker
 *  when attempting to load shared libraries)
 */
#ifndef PATHMAPFILE
/* Remap just this internal wired set of names */
int
lx_pathname(const char *lx_path, char *os_path, int len)
{
        int             n = getstring(lx_path, os_path, len);
        int             r;

        if (n < 0)
        {
	    return enoany(EFAULT, "pathname() - getstring fail");
        }

        if (n == len && os_path[n-1] != '\0')
        {
	    return enoany(ENAMETOOLONG, "pathname() - os_path truncated");
        }
                                
        /*
         * simple hack for now - just trap accesses to /lib and /usr/lib
         */
                        
        if (linux_root && (strncmp(os_path, "/lib/", 5) == 0 ||
                strncmp(os_path, "/usr/lib/", 9) == 0 ||
                strncmp(os_path, "/usr/local/lib/", 15) == 0 ||
                strncmp(os_path, "/usr/X11R6/lib/", 15) == 0) ) 
        {
        
                r = strlen(linux_root);
                if (r + n > len) {
		    return enoany(ENAMETOOLONG,
				  "pathname() - map name too long");
                }               
                memmove(os_path+r, os_path, n);
                memcpy(os_path, linux_root, r);
           
                return n + r;
        }
                
        return n;

}

#else   /* defined(PATHMAPFILE) */

/* 
 * Use set of file maps from external file read into an internal list
 */
int
lx_pathname(const char *lx_path, char *os_path, int len)
{
    int		n;
    int		r;
    pathmap_t	*pm;
    char	*soptr;
    struct stat	sbuf;

    PATHMAP_TRACE_MSG(("TRACE: lx_pathname(\"%s\", %#x, %d)\n",
					     lx_path, os_path, len));

    if ((n = getstring(lx_path, os_path, len)) < 0)
    {
	    return enoany(EFAULT, "pathname() - getstring fail");
    }

    if ((n == len) && (os_path[n-1] != '\0'))
    {
	return enoany(ENAMETOOLONG, "pathname() - os_path truncated");
    }

    /* If the input name starts with the root simply return.
     * - this is an optimisation otherwise handled in the PATHMAP file
     */
    {
	size_t lrlen = strlen(linux_root);
		
	if (strncmp(lx_path, linux_root, lrlen) == 0)
	{
	    (void) strcpy(os_path, lx_path);
	    return len;
	}
    }

    strncpy(os_path, lx_path, len);
    /*
     * Traverse linked list of pathname maps and apply the
     * first appropriate one we find.
     */

    for (pm = pathmap_head; pm; pm = pm->next)
	if ((strncmp(pm->from, lx_path, pm->fromlen) == 0) &&
	    (!lx_path[pm->fromlen] || (lx_path[pm->fromlen] == '/')))
	{
	    switch (pm->type)
	    {
		case PREFIX:
		    strncpy(os_path, linux_root, len);
		    if (lx_path[0] != '/')
			strncat(os_path, "/", len);
		    strncat(os_path, lx_path, len);
		    break;

		case DOMAP:
		    strncpy(os_path, pm->to, len);
		    if (lx_path[pm->fromlen])
			strncat(os_path, &lx_path[pm->fromlen], len);
		    break;

#ifdef PHANTOM_MAP_FILE
		case DOEXEC:
		   return create_phantom_file(pm, lx_path, os_path, len);
		   break;
#endif                    
		case NOMAP:
		    /* Nothing to do here */
		    break;
	    }

	    break;
	}

#ifdef BREAKS_GIMP	/* Why do we do this? */
    /*
     * Check if this is might be a library, and if it doesn't exist,
     * try the library name without the .so version
     */

    if ((soptr = strstr(os_path, ".so.")) != NULL)
  	if (stat(os_path, &sbuf) && (errno == ENOENT))
    {
	
	PATHMAP_TRACE_MSG(("TRACE: lx_pathname: shobj \"%s\" not found\n",
			   os_path));
	soptr[3] = '\0';
    }
#endif
    

#if 0		/* This is just too ugly.  It should be handled elsewhere. */
    /*
     * map /proc/self onto /proc/<pid>
     */

    if ((strncmp(os_path, "/proc/self", 10) == 0) &&
	(!os_path[10] || (os_path[10] == '/')))
    {
	char	tmp[PATH_MAX];

	strcpy(tmp, &os_path[10]);
	sprintf(os_path, "/proc/%d%s", getpid(), tmp);
    }
#endif

    PATHMAP_TRACE_MSG(("TRACE: lx_pathname: mapped to \"%s\"\n", os_path));
    return strlen(os_path);
}

/*
 * Called once on startup to parse pathmap file and setup lxrun
 * internal pathname map structure.
 */
int
parse_pathmap_table(const char *fn)
{
    FILE        *fp;
    char	buf[256];
    char	*s;
    const char	*from;
    const char	*to;
    int         linenum;
    pathmap_t	*mp;
    pathmap_t	**mpp;
    char	meta_from[PATH_MAX];
    char	meta_to[PATH_MAX];

    /* open pathmap definition file */
    if((fp = fopen(fn, "r")) == NULL)
		return -1;

    if (!linux_root || !*linux_root)
	linux_root = "/";

    for (linenum = 1; (s = fgets(buf, sizeof buf, fp)) != NULL; linenum++)
    {
	while (isspace(*s))
	    ++s;			/* skip leading whitespace */

	if (!*s ||			/* Ignore blank and	*/
#ifndef TRACE
	    (*s == ':') ||		/* PATHMAP TRACE testing */
#endif
	    (*s == '#') || (*s == ';'))	/*   comment lines	*/
			continue;

	from = s;

	while (*s && !isspace(*s))
	    s++;			/* Find end of from */

        /*
         * assert: *s is the first non-whitespace character
         * in a non-comment line.
         */
        if (*s)
	{
	    *s = '\0';			/* End of from */
	    if ((&s[-1] > from) && (s[-1] == '/'))
		s[-1] = '\0';		/* Remove trailing '/' */
	    ++s;
	}

	while (isspace(*s))
	    ++s;			/* Find beginning of to */

	if (!*s)
        {
	    warning(0, "%s: line %d: `to' field expected\n", fn, linenum);
	    continue;
	}

	/* assert: *s is the first character in the "to" field. */
	to = s;

	while (*s && !isspace(*s))
	    ++s;			/* Find end of to */

	*s='\0';

	if ((from[0] == ':') && (do_trace & TRACE_PATHMAP))
	{
	    if (strncmp(to, "+/", 2) == 0)
	    {
		sprintf(meta_from, "%s/%s", linux_root, &to[2]);
		to = meta_from;
	    }

	    if (lx_pathname(to, meta_to, sizeof(buf)) == -1)
		tprintf("TRACE: Map test fail: %s\n", meta_to);
	    else
		tprintf("TRACE: Map test: %s\n\t     --> %s\n", to, meta_to);

	    continue;
	}

	/*
	 * If we made it this far, then from and to point to
	 * nul-terminated strings each containing at least one
	 * character and neither containing any whitespace.
	 *
	 * Note: memory for the maps is never deallocated.
	 */

	if (!(mp = (pathmap_t *)malloc(sizeof(pathmap_t))))
	    fatal(ENOMEM, "Cannot malloc for new map\n");

	memset(mp, 0, sizeof(*mp));

	/*
	 * Fixup meta-path in from
	 */

	if (strcmp(from, "+") == 0)
	    from = linux_root;
	else if (strncmp(from, "+/", 2) == 0)
	{
	    sprintf(meta_from, "%s/%s", linux_root, &from[2]);
	    from = meta_from;
	}

	if (!(mp->from = strdup(from)))
	    fatal(ENOMEM, "Cannot malloc for new from-map\n");

	mp->fromlen = strlen(mp->from);

	/*
	 * Setup operation type
	 */

        mp->type = emt_UNDEF;       /* appear as BOGUS in trace file */
	if (strcmp(to, "-") == 0)
	    mp->type = NOMAP;
	else if (strcmp(to, "+") == 0)
	    mp->type = PREFIX;
	else
	{
	    if (*to == '+' )
                mp->type = DOMAP;
#ifdef PHANTOM_MAP_FILE
	    if (*to == '@' )
	        mp->type = DOEXEC;
#endif
	    /*
	     * Fixup meta-path in to
	     */

	    if (mp->type != emt_UNDEF && to[1] == '/')
	    {
		sprintf(meta_to, "%s/%s", linux_root, &to[2]);
		to = meta_to;
	    }

	    if (!(mp->to = strdup(to)))
		fatal(ENOMEM, "Cannot malloc for new to-map\n");
	}

	/*
	 * Insert the new entry on the end of the linked list.
	 */

	for (mpp = &pathmap_head; *mpp; mpp = &(*mpp)->next)
	    ;

	*mpp = mp;

	PATHMAP_TRACE_MSG(("TRACE: New %s map: %s%s%s\n",
					    map_type_str(mp->type),
					    mp->from,
					    mp->to ? " --> " : "",
					    mp->to ? mp->to : ""));
    }

    fclose(fp);
    return 0;
}

#if defined(PHANTOM_MAP_FILE)
/* 
 * Supports a special map that is an external program to run to 
 * initialise a (temporary) file with the info required 
 */

#include        <sys/types.h>
#include        <sys/wait.h>
#include        <fcntl.h>

static char	phantom_file[20];
static char	phantom_file_valid = 0;



/*
 * Create a phantom file based on the information in pm.
 * Return its name in fn, unless it's longer than len.
 * If the name is too long, return -1 and set errno to ENAMETOOLONG.
 */
static int
create_phantom_file(pathmap_t *pm, const char *from, char *fn, int len)
{
	int fd, phlen;
	char fd_str[10];

	/*
	** It's okay that the name is always the same for this process
	** since we delete it immediately anyway.  If we need more than
	** one phantom file per system call, then we're in trouble, and
	** this mechanism will need to be changed.
	*/

        PATHMAP_TRACE_MSG(("TRACE: Calling %s to generate %s\n", pm->to,from));

	snprintf(phantom_file, sizeof(phantom_file), "/tmp/lx%d", getpid());

	if((phlen = strlen(phantom_file))+1 > len)
	{
	    return enoany(ENAMETOOLONG, "TRACE: phantom filename too long");
	}

	/* Create the phantom file */
	if((fd=open(phantom_file, O_RDWR | O_CREAT, 0600))==-1)
	{
	    PATHMAP_TRACE_MSG(("TRACE: Creation %s FAILED\n", from));
	    return -1;
	}
	
	phantom_file_valid=1;
	sprintf(fd_str, "%d", fd);

	/* Send the name of the phantom file back to the Linux process */
	strcpy(fn, phantom_file);

	/* Call our executable to write out the file */
	if (fork() == 0)
	{
		char buf[512];

		/* cmdline; truncate if necessary */
		strcpy(buf, "LINUX_CMDLINE=");
		strncpy(buf+14, linux_cmdline, (sizeof buf)-15);
		buf[sizeof buf - 1] = '\0';
		putenv(strdup(buf));

		/* our pid */
		sprintf(buf, "LINUX_PID=%d", getpid());
		putenv(strdup(buf));

		/*
		** Write the phantom file represeting "from" to
		** file descriptor "fd".
		*/
		execl(pm->to, pm->to, fd_str, from, NULL);
		/* We should print out an error message here, maybe. */
		exit(-1);
	}

	/* The parent doesn't need the phantom file fd */
	close(fd);

	/* Wait for the child to write the phantom file. */
	wait((int *)0);

	PATHMAP_TRACE_MSG(("TRACE: %s execution done\n", pm->to));
	return phlen;
}


/* Remove any phantom file */
void remove_phantom_file(void)
{
	if (phantom_file_valid)
	{
	    unlink(phantom_file);
	    phantom_file_valid = 0;
	}
}

#endif  /* PHANTOM_MAP_FILE */
#endif  /* PATHMAPFILE */



static const char *
map_type_str(map_t type)
{
    switch (type)
    {
	case NOMAP:
	    return "NOMAP";

	case PREFIX:
	    return "PREFIX";

#ifdef PHANTOM_MAP_FILE
	case DOEXEC:
	   return "DOEXEC";
	   break;
#endif                    
	case DOMAP:
	    return "DOMAP";
    }

    return "BOGUS";
}


/*
 * if linux progname is not a pathname search for it in the PATH
 *  and update it to its absolute pathname.
 * 	if  LXRUN_AUTO_PATH_BEHAVIOR defined
 */
void
chkpath_lxprog(const char *pathenv, char **lxprog)
{
#ifdef  LXRUN_AUTO_PATH_BEHAVIOR
    static char	qualified[PATH_MAX];
    /* We use a copy of pathenv since strtok is destructive. */
    char	*pathcopy = strdup(pathenv);
    char	*path = pathcopy;
		
    if (strchr(*lxprog, '/'))	/* progname contains slash - assume OK as is */
	return;
   
    /* check for linux prog in our PATH */

    while((path = strtok(path, ":")))
    {
	strncpy(qualified, path, sizeof(qualified));
	if (qualified[0] == '\0')
	    strncpy( qualified, ".", sizeof(qualified));
	strncat(qualified, "/",     sizeof(qualified));
	strncat(qualified, *lxprog, sizeof(qualified));

	/* Lxrun can't load the binary w/o read permissions. */
	if( EACCESS(qualified, R_OK) &&  EACCESS(qualified, X_OK))
	{
	    *lxprog = qualified;
	    break;
	}

	path = NULL;	/* reset path ptr for strtok() */
    }
    free(pathcopy);
#endif    
}

