#ifdef OSR5
#pragma comment(exestr, "%Z% %M% %I% %E% %Q%")
#endif

#ifdef Solaris
#pragma ident	"%Z%%M%	%I%	%E% SMI"
#endif
/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<errno.h>
#include	<fcntl.h>
#include	<limits.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<sys/mman.h>
#include	<sys/time.h>
#include	<sys/resource.h>

#include	"lxrun.h"
#include	"linux/types.h"
#include	"linux/resource.h"
#include	"linux/sched.h"
#include	"trace.h"

char	emulator[PATH_MAX] = EMULATOR;

extern caddr_t	minimum_brk;
extern caddr_t	current_brk;

extern void	*dz_mmap(void *, size_t, int, int, off_t);

/*
 * process control system calls
 *
 *	execve
 *	ptrace		(NOT IMPLEMENTED)
 *	brk
 *	setrlimit
 *	getrlimit
 *	getrusage
 *	getgroups
 *	setgroups
 *	getpriority	(NOT IMPLEMENTED)
 *	setpriority	(NOT IMPLEMENTED)
 *	uselib
 *	personality
 *      sched_setparam	  (NOT IMPLEMENTED)
 *      sched_getparam	  (hackish stub)
 *      sched_setscheduler      (hackish stub)
 *      sched_getscheduler      (hackish stub)
 *      sched_yield	        (Solaris only, hackish stub for others)
 *      sched_get_priority_max  (hackish stub)
 *      sched_get_priority_min  (hackish stub)
 *      sched_rr_get_interval   (NOT IMPLEMENTED)
 */


/*
 * Nuke some environment variables. In particular, netscape likes to
 * LD_PRELOAD some linux libraries (/lib/libBrokenLocale.so.1) before
 * execing the netscape binary, but that invokes lxrun, and our ld doesn't
 * run lxrun because it can't find those libraries.
 */
static struct envtable {
	char	*var;
	size_t	len;
} envtable [] = {
	{ "LD_PRELOAD=", sizeof ("LD_PRELOAD=") - 1},
	{ NULL, 0 }
};

static void
clear_ld_environment(char **envp)
{
	struct envtable *etbl;
	for (; *envp != NULL; envp++) {
		for (etbl = &envtable[0]; etbl->len != 0; etbl++) {
			if (strncmp(*envp, etbl->var, etbl->len) == 0)
				(*envp)[etbl->len] = '\0';
		}
	}
}


int
lx_execve(char *prog, char **argv, char **envp)
{
	int  ac = 0;
	char **av;
	char real_interp[PATH_MAX];
	char interp_arg[PATH_MAX];

	/*
	 * count args, including terminating NULL
	 */
	TRACE_MSG( TRACE_EXEC, ("TRACE: lx_execve(\"%s\"", prog));
	while (argv[ac++])
	    TRACE_MSG( TRACE_EXEC, (", \"%s\"", argv[ac - 1]));
	TRACE_MSG( TRACE_EXEC, (")\n"));

	clear_ld_environment(envp);

	if (emulator[0] == '\0')
	{
		if (is_interp_script(prog, real_interp, sizeof (real_interp),
				interp_arg, sizeof (interp_arg)))
		{
		    av = alloca((ac + 2) * sizeof(char *));
		    if (!av)
		      return enoany(ENOMEM, "TRACE: lx_execve: alloca failed");

		    av[0] = argv[0];
		    if (interp_arg[0] != '\0')
			{
			av[1] = interp_arg;
			av[2] = prog;
			memcpy(&av[3], &argv[1], (ac - 1) * sizeof (char *));
			}
		    else
			{
			av[1] = prog;
			memcpy(&av[2], &argv[1], (ac - 1) * sizeof (char *));
			}

		    prog = real_interp;
		    argv = av;
		}

	}
	else if (is_linux_exec(prog))
	{
		TRACE_MSG(TRACE_EXEC,
			("TRACE: is_linux_exec(%s)=true\n", prog));
		/*
		 * Allocate new argv and copy old args.
		 * Use the "-0" flag to pass in the desired argv[0].
		 */
		av = alloca((ac + 3) * sizeof(char *));
		if (!av)
		    return enoany(ENOMEM, "TRACE: lx_execve: alloca failed");

		av[0] = emulator;
		av[1] = "-0";
		av[2] = argv[0];
		av[3] = prog;
		memcpy(&av[4], &argv[1], (ac - 1) * sizeof (char *));

		prog = emulator;
		argv = av;
	}
	else if (is_interp_script(prog, real_interp, sizeof (real_interp),
				interp_arg, sizeof (interp_arg)))
	{
		TRACE_MSG(TRACE_EXEC,
			("TRACE: is_interp_script(%s)=true\n", prog));
		/*
		 * Feed the script to the interpreter.
		 */
		av = alloca((ac + 5) * sizeof(char *));
		if (!av)
		    return enoany(ENOMEM, "TRACE: lx_execve: alloca failed");

		av[0] = emulator;
		av[1] = "-0";
		av[2] = argv[0];
		av[3] = real_interp;
		if (interp_arg[0] != '\0')
			{
			av[4] = interp_arg;
			av[5] = prog;
			memcpy(&av[6], &argv[1], (ac - 1) * sizeof (char *));
			}
		else
			{
			av[4] = prog;
			memcpy(&av[5], &argv[1], (ac - 1) * sizeof (char *));
			}

		prog = emulator;
		argv = av;
	}

	/* Show actual call to execve */
	TRACE_MSG( TRACE_EXEC, ("TRACE: execve(\"%s\"", prog));
	for(ac=0; argv[ac]; ac++)
	    TRACE_MSG( TRACE_EXEC, (", \"%s\"", argv[ac]));
	TRACE_MSG( TRACE_EXEC, (")\n"));

	return execve(prog, argv, envp);
}

int lx_ptrace()		{ return enosys("ptrace"); }

void *
lx_brk(caddr_t brkval)
{
	unsigned long	new_brk;
	unsigned long	old_brk;
	void		*ret;

	TRACE_MSG(TRACE_BRK, ("TRACE: lx_brk(0x%08x)\n", brkval));

	if (brkval < minimum_brk)
		return current_brk;

	new_brk = PAGE_ALIGN(brkval);
	old_brk	= PAGE_ALIGN(current_brk);

	if (new_brk == old_brk)
	    return current_brk = brkval;

	if (new_brk < old_brk)
	{
		munmap((caddr_t)new_brk, old_brk - new_brk);
		return current_brk = brkval;
	}

	ret = dz_mmap((caddr_t)old_brk, new_brk - old_brk,
		      PROT_READ|PROT_WRITE|PROT_EXEC,
		      MAP_FIXED|MAP_PRIVATE, 0);
	if (ret == (caddr_t)old_brk)
	{
	    current_brk = brkval;
	}
	else if (ret == (caddr_t)-1)	/* MAP_FAILED */
	{
	    TRACE_MSG(TRACE_BRK,
		       ("TRACE: lx_brk(%#x), dz_mmap failed: errno %d\n",
			brkval, errno));
	    return ret;
	}

	return current_brk;
}

int
lx_setrlimit(int resource, const struct rlimit *rlp)	/* S.G. */
{
	int rsrc;

	switch(resource)
	{
	/* Cases that map directly from Linux to SCO */
	case LX_RLIMIT_CPU:	rsrc = RLIMIT_CPU;    break;
	case LX_RLIMIT_FSIZE:	rsrc = RLIMIT_FSIZE;  break;
	case LX_RLIMIT_DATA:	rsrc = RLIMIT_DATA;   break;
	case LX_RLIMIT_STACK:	rsrc = RLIMIT_STACK;  break;
	case LX_RLIMIT_CORE:	rsrc = RLIMIT_CORE;   break;
	case LX_RLIMIT_NOFILE:	rsrc = RLIMIT_NOFILE; break;

	/* Cases that don't.  :-(  */
	case LX_RLIMIT_RSS:
	    return enosys("setrlimit(RLIMIT_RSS)");
	case LX_RLIMIT_NPROC:
	    return enosys("setrlimit(RLIMIT_NPROC)");
	default:
		return enosys("setrlimit(?)");
	}

    	return setrlimit(rsrc, rlp);
}


int
lx_getrlimit(int resource, struct rlimit *rlp)
{
    	int rsrc;

    	switch(resource)
	{
	/* Cases that map directly from Linux to SCO */
	case LX_RLIMIT_CPU:	rsrc = RLIMIT_CPU;    break;
	case LX_RLIMIT_FSIZE:	rsrc = RLIMIT_FSIZE;  break;
	case LX_RLIMIT_DATA:	rsrc = RLIMIT_DATA;   break;
	case LX_RLIMIT_STACK:	rsrc = RLIMIT_STACK;  break;
	case LX_RLIMIT_CORE:	rsrc = RLIMIT_CORE;   break;
	case LX_RLIMIT_NOFILE:	rsrc = RLIMIT_NOFILE; break;

	/* Cases that don't.  :-(  */
	case LX_RLIMIT_RSS:
	    return enosys("getrlimit(RLIMIT_RSS)"); 
	    
	case LX_RLIMIT_NPROC:
	    return enosys("getrlimit(RLIMIT_NPROC)");
	   
	default:
	    return enosys("getrlimit(?)");
	}

    	return getrlimit(rsrc, rlp);
}


/*
 * Solaris and Linux match here, except Linux has RUSAGE_BOTH
 */
#define RUSAGE_BOTH	-2		/* sys_wait4() uses this */

int
lx_getrusage(int who, struct rusage *ru)
{
#if defined(RUSAGE_SELF)
	int rval;

	struct rusage rc;

	switch (who) {
	case RUSAGE_SELF:
	case RUSAGE_CHILDREN:
		rval = getrusage(who, ru);
		break;
	case RUSAGE_BOTH:
		rval = getrusage(RUSAGE_SELF, ru);
		if (rval != -1) {
			rval = getrusage(RUSAGE_CHILDREN, &rc);
			if (rval != -1) {
				ru->ru_utime.tv_sec += rc.ru_utime.tv_sec;
				ru->ru_utime.tv_usec += rc.ru_utime.tv_usec;
				ru->ru_stime.tv_sec += rc.ru_stime.tv_sec;
				ru->ru_stime.tv_usec += rc.ru_stime.tv_usec;
				ru->ru_maxrss += rc.ru_maxrss;
				ru->ru_ixrss += rc.ru_ixrss;
				ru->ru_idrss += rc.ru_idrss;
				ru->ru_isrss += rc.ru_isrss;
				ru->ru_minflt += rc.ru_minflt;
				ru->ru_majflt += rc.ru_majflt;
				ru->ru_nswap += rc.ru_nswap;
				ru->ru_inblock += rc.ru_inblock;
				ru->ru_oublock += rc.ru_oublock;
				ru->ru_msgsnd += rc.ru_msgsnd;
				ru->ru_msgrcv += rc.ru_msgrcv;
				ru->ru_nsignals += rc.ru_nsignals;
				ru->ru_nvcsw += rc.ru_nvcsw;
				ru->ru_nivcsw += rc.ru_nivcsw;
			}
		}
		break;
	default:
		errno = EINVAL;
		return (-1);
	}
	return (rval);
#else
	return enosys("getrusage");
#endif
}


int
lx_getgroups(int gidsetsize, lx_gid_t *lx_grouplist)
{
	gid_t	grouplist[NGROUPS_MAX];
	int	r;

	if (gidsetsize < 0 || gidsetsize > NGROUPS_MAX)
	{
		return enoany(EINVAL, "getgroups(! 0<gidsetsize>NGROUPS_MAX)");
	}


	r = getgroups(gidsetsize, grouplist);

	if (r > 0 && gidsetsize > 0)
	{
		int	i;

		if (! verify_write(lx_grouplist, r * sizeof(lx_grouplist[0])))
			return -1;

		for (i = 0; i < r; i++)
			lx_grouplist[i] = (lx_gid_t) grouplist[i];
	}

	return r;
}	


int
lx_setgroups(int gidsetsize, lx_gid_t *lx_grouplist)
{
	gid_t	grouplist[NGROUPS_MAX];
	int	i;

	if (gidsetsize < 0 || gidsetsize > NGROUPS_MAX)
	{
		return enoany(EINVAL, "setgroups(! 0<gidsetsize>NGROUPS_MAX)");
	}

	if (! verify_read(lx_grouplist, gidsetsize * sizeof(lx_grouplist[0])))
		return -1;

	for (i = 0; i < gidsetsize; i++)
		grouplist[i] = (gid_t) lx_grouplist[i];

	return setgroups(gidsetsize, grouplist);
}


int lx_getpriority()	{ return enosys("getpriority"); }
int lx_setpriority()	{ return enosys("setpriority"); }

int
lx_uselib(char *libname)
{
	if (load_library(libname) != NULL)
		return 0;

	return -1;
}	


int
lx_personality(unsigned long personality)
{
	if (personality == 0xffffffff || personality == 0)
		return 0;
		
	return enoany(EINVAL, "personality(!0L or !~0L)");
}


int
lx_sched_yield(void)
{
#ifdef SOLARIS
	yield();
#endif
	return 0;
}


int
lx_sched_get_priority_max(int policy)
{
	switch(policy) {
	    case LX_SCHED_FIFO:
	    case LX_SCHED_RR:
		return 99;
	    case LX_SCHED_OTHER:
		return 0;
	}
	errno = EINVAL;
	return -1;
}


int
lx_sched_get_priority_min(int policy)
{
	switch(policy) {
	    case LX_SCHED_FIFO:
	    case LX_SCHED_RR:
		return 1;
	    case LX_SCHED_OTHER:
		return 0;
	}
	errno = EINVAL;
	return -1;
}


int
lx_sched_getscheduler(pid_t pid)
{
	return LX_SCHED_OTHER;
}

int
lx_sched_setscheduler(pid_t pid, int policy, struct lx_sched_param *param)
{
	return 0;
}

int
lx_sched_getparam(pid_t pid, struct lx_sched_param *param) 
{
	param->sched_priority = 0;
	return 0;
}

int lx_sched_setparam()		{ return enosys("sched_setparam"); }
int lx_sched_rr_get_interval()	{ return enosys("sched_rr_get_interval"); }



/* ---------------------------------------------------------------- */
/* We don't want anything in lxrun to call brk().
 * Replace malloc for ld-lxrun.so */
/* ++ should be renamed
 * (temporary hack: we have -Dmalloc=lx_malloc, etc. in Makefile) */
/* Should this be in a separate source file? */

#include <stdio.h>
#include <memory.h>
#include <malloc.h>

static char	malloc_pool[128 * 1024];
static char	*malloc_ptr = malloc_pool;

void *
malloc(size_t size)
{
    void	*newptr;

    if ((malloc_ptr + size) >= &malloc_pool[sizeof(malloc_pool)])
    {
	write(fileno(stderr), "Out of memory\n", 14);
	exit(ENOMEM);
    }

    newptr = malloc_ptr;
    malloc_ptr += size;
    return newptr;
}


char *
strdup(const char *str)
{
    void	*newstr;
    newstr = (char *)malloc(strlen(str)+1);
    strcpy(newstr, str);
    return newstr;
}


void *
realloc(void *ptr, size_t size)
{
    void	*newptr;

    if (!(newptr = malloc(size)))
	return NULL;

    if (ptr != NULL) {
	memcpy(newptr, ptr, size);
	free(ptr);
    }
    return newptr;
}

void
free(void *ptr)
{
    /* ## */
}

#ifdef NEED_CALLOC

void *
calloc(size_t nelem, size_t elsize)
{
    void	*newptr;
    u_long	size = nelem * elsize;

    newptr = malloc(size);
    if (newptr)
	memset(newptr, 0, size);
    return newptr;
}

#endif	/* NEED_CALLOC */

