#ifdef OSR5
#pragma comment(exestr, "%Z% %M% %I% %E% %Q%")
#endif
/*
 * Copyright (C) 1998,1999 SCO Inc
 * All rights reserved.
 *
 * This software is provided "as is" without express or implied warranty.
 */

/*
 * The following have been considered but not handled:
 *	dup()
 *	fork()/exec() with /proc stuff open.
 *	readdir() and getdents()
 *
 * /proc/self is mapped to /proc/<pid> by lx_pathname().
 */

#ifdef OLD_PROCFS

#include	<stdio.h>
#include	<sys/stat.h>
#include	<sys/uio.h>
#include	<fcntl.h>
#include	<signal.h>
#include	<memory.h>
#include	<malloc.h>
#include	<errno.h>
#include	<unistd.h>
#include	<ctype.h>
#include	<string.h>
#include	<unistd.h>
#include	<time.h>

#ifdef UW7
#   include	<sys/procfs.h>
#   include	<sys/fault.h>
#   include	<sys/fault.h>
#   include	<sys/syscall.h>
#   include	<macros.h>
#endif

#include	"lxrun.h"
#include	"procfs.h"
#include	"trace.h"

typedef struct
{
    const char	*name;
    int		(*open)(int pid, int flags, int mode);
    int		(*stat)(int pid, struct stat *buf);
    int		(*lstat)(int pid, struct stat *buf);
    int		(*access)(int pid, int mode);
    int		(*read)(int pid, int fd, void *buf, int len);
    int		(*write)(int pid, int fd, void *buf, int len);
    int		(*ioctl)(int pid, int fd, int cmd, unsigned long arg);
    int		(*fcntl)(int pid, int fd, int cmd, int arg);
} lxprocfs_t;

typedef struct procfssw_s procfssw_t;
struct procfssw_s
{
    int		fd;
    int		pid;
    lxprocfs_t	*op;
    procfssw_t	*next;
};

static procfssw_t	*procfs = NULL;
static procfssw_t	*procfs_spares = NULL;

static procfssw_t *find_procfssw(int fd);
static lxprocfs_t *lxprocfs_ops(const char *path, int *pidp);

int
lxprocfs_stat(const char *path, struct stat *buf)
{
    lxprocfs_t	*op;
    int		pid;

    if (!(op = lxprocfs_ops(path, &pid)))
	return -1;	/* Use errno of previous stat() call */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_stat(%s) %s\n", path, op->name));
    return op->stat(pid, buf);
}

int
lxprocfs_lstat(const char *path, struct stat *buf)
{
    lxprocfs_t	*op;
    int		pid;

    if (!(op = lxprocfs_ops(path, &pid)))
	return -1;	/* Use errno of previous lstat() call */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_lstat(%s) %s\n", path, op->name));
    return op->lstat(pid, buf);
}

int
lxprocfs_fstat(int fd, struct stat *buf)
{
    procfssw_t	*fp;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Use errno of previous fstat() call */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_fstat(%d) %s\n", fd, fp->op->name));
    return fp->op->stat(fp->pid, buf);
}

int
lxprocfs_access(const char *path, int mode)
{
    lxprocfs_t	*op;
    int		pid;

    if (!(op = lxprocfs_ops(path, &pid)))
	return -1;	/* Use errno of previous access() call */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_access(%s) %s\n", path, op->name));
    return op->access(pid, mode);
}

int
lxprocfs_open(const char *path, int flags, int mode)
{
    lxprocfs_t	*op;
    int		pid;
    int		fd;
    procfssw_t	*fp;
    procfssw_t	**fpp;

    if (!(op = lxprocfs_ops(path, &pid)))
	return -1;	/* Use errno of previous open() call */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_open(%s) %s\n", path, op->name));

    if ((fd = op->open(pid, flags, mode)) == -1)
	return -1;	/* op->open() provided errno */

    /*
     * Our malloc pool is a bit soft.  We have a structure. 
     * Keep it for next time.
     */

    for (fpp = &procfs; *fpp; fpp = &(*fpp)->next)
	;

    if (procfs_spares)
    {
	fp = procfs_spares;
	procfs_spares = fp->next;
	memset(fp, 0, sizeof(*fp));
    }
    else if (!(fp = (procfssw_t *)malloc(sizeof(*fp))))
    {
	close(fd);
	return enoany(ENOMEM, "procfs_open() - malloc fail");
    }

    *fpp = fp;
    fp->fd = fd;
    fp->pid = pid;
    fp->op = op;

    return fd;
}

int
lxprocfs_close(int fd)
{
    procfssw_t	*fp;
    procfssw_t	**fpp;

    for (fpp = &procfs; (fp = *fpp) != NULL; fpp = &fp->next)
	if (fp->fd == fd)
	{
	    PROCFS_TRACE_MSG(("TRACE: lxprocfs_close(%d)\n", fd, fp->op->name));
	    *fpp = fp->next;
	    fp->next = procfs_spares;
	    procfs_spares = fp;
	    close(fd);
	    return 0;
	}

    return -1;	/* Not ours.  Let subsequent close() set errno */
}

int
lxprocfs_read(int fd, void *buf, int len)
{
    procfssw_t	*fp;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Not ours.  Let subsequent read() set errno */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_read(%d) %s\n", fd, fp->op->name));
    return fp->op->read(fp->pid, fd, buf, len);
}

int
lxprocfs_readv(int fd, struct iovec *iov, int iovcnt)
{
    procfssw_t	*fp;
    int		cnt;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Not ours.  Let subsequent read() set errno */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_readv(%d) %s\n", fd, fp->op->name));

    cnt = 0;
    while (iovcnt)
    {
	int	err;

	err = fp->op->read(fp->pid, fd, iov->iov_base, iov->iov_len);
	if (err == -1)
	    return -1;

	cnt += err;
	if (err |= iov->iov_len)
	    break;

	++iov;
	--iovcnt;
    }

    return cnt;
}

int
lxprocfs_write(int fd, void *buf, int len)
{
    procfssw_t	*fp;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Not ours.  Let subsequent write() set errno */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_write(%d) %s\n", fd, fp->op->name));
    return fp->op->write(fp->pid, fd, buf, len);
}

int
lxprocfs_writev(int fd, const struct iovec *iov, int iovcnt)
{
    procfssw_t	*fp;
    int		cnt;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Not ours.  Let subsequent writev() set errno */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_writev(%d) %s\n", fd, fp->op->name));

    cnt = 0;
    while (iovcnt)
    {
	int	err;

	err = fp->op->write(fp->pid, fd, iov->iov_base, iov->iov_len);
	if (err == -1)
	    return -1;

	cnt += err;
	if (err |= iov->iov_len)
	    break;

	++iov;
	--iovcnt;
    }

    return cnt;
}

int
lxprocfs_ioctl(int fd, int cmd, unsigned long arg)
{
    procfssw_t	*fp;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Not ours.  Let subsequent ioctl() set errno */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_ioctl(%d) %s\n", fd, fp->op->name));
    return fp->op->ioctl(fp->pid, fd, cmd, arg);
}

int
lxprocfs_fcntl(int fd, int cmd, int arg)
{
    procfssw_t	*fp;

    if (!(fp = find_procfssw(fd)))
	return -1;	/* Not ours.  Let subsequent fcntl() set errno */

    PROCFS_TRACE_MSG(("TRACE: lxprocfs_fcntl(%d) %s\n", fd, fp->op->name));
    return fp->op->fcntl(fp->pid, fd, cmd, arg);
}

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									 *
 *				nodev stubs				 *
 *									 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * The intent of these routines is to provide stubs for
 * functions that do not work on some device while providing
 * some trace messages.  Notice that there is no
 * lxp_nodev_open().  Without an open there is no implementation.
 */

static int lxp_nodev_stat(int pid, struct stat *buf);
static int lxp_nodev_access(int pid, int mode);
static int lxp_nodev_read(int pid, int fd, void *buf, int len);
static int lxp_nodev_write(int pid, int fd, void *buf, int len);
static int lxp_nodev_ioctl(int pid, int fd, int cmd, unsigned long arg);
static int lxp_nodev_fcntl(int pid, int fd, int cmd, int arg);

static int
lxp_nodev_stat(int pid, struct stat *buf)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_nodev_stat()\n"));

    return enoany(ENOSYS, "");
}

static int
lxp_nodev_access(int pid, int mode)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_nodev_access()\n"));

    return enoany(ENOSYS, "");
}

static int
lxp_nodev_read(int pid, int fd, void *buf, int len)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_nodev_read()\n"));

    return enoany(ENOSYS, "");
}

static int
lxp_nodev_write(int pid, int fd, void *buf, int len)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_nodev_write()\n"));

    return enoany(ENOSYS, "");
}

static int
lxp_nodev_ioctl(int pid, int fd, int cmd, unsigned long arg)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_nodev_ioctl()\n"));

    return enoany(ENOSYS, "");
}

static int
lxp_nodev_fcntl(int pid, int fd, int cmd, int arg)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_nodev_fcntl()\n"));

    return enoany(ENOSYS, "");
}

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									 *
 *			    The directory /proc				 *
 *									 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

static int lxp_proc_open(int pid, int flags, int mode);
static int lxp_proc_stat(int pid, struct stat *buf);
static int lxp_proc_access(int pid, int mode);
static int lxp_proc_read(int pid, int fd, void *buf, int len);

static const char	procName[] = "/proc";

static lxprocfs_t	lxp_proc_ops =
{
    procName,
    lxp_proc_open,
    lxp_proc_stat,
    lxp_proc_stat,	/* We are a directory stat() and lstat() are same */
    lxp_proc_access,
    lxp_proc_read,
    lxp_nodev_write,
    lxp_nodev_ioctl,
    lxp_nodev_fcntl
};

/*
 * Here we will simulate /proc with the help of the real /proc
 *
 * ## OpenServer problem here!!
 */

static int
lxp_proc_open(int pid, int flags, int mode)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_proc_open(%s)\n", procName));
    return open(procName, flags, mode);
}

static int
lxp_proc_stat(int pid, struct stat *buf)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_proc_stat(%s)\n", procName));
    return stat(procName, buf);
}

static int
lxp_proc_access(int pid, int mode)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_proc_access(%s)\n", procName));
    return access(procName, mode);
}

static int
lxp_proc_read(int pid, int fd, void *buf, int len)
{
    int		r;

    PROCFS_TRACE_MSG(("TRACE: lxp_proc_read(%s)\n", procName));

    if ((r = read(fd, buf, len)) != -1)
	return r;

    /*
     * ## Need code here for simulating an entry for /proc/self
     */

    return -1;
}

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									 *
 *			The directory /proc/<pid>			 *
 *									 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

static int lxp_procpid_open(int pid, int flags, int mode);
static int lxp_procpid_stat(int pid, struct stat *buf);
static int lxp_procpid_access(int pid, int mode);
static int lxp_procpid_read(int pid, int fd, void *buf, int len);

static lxprocfs_t	lxp_procpid_ops =
{
    "procpid",
    lxp_procpid_open,
    lxp_procpid_stat,
    lxp_procpid_stat,	/* We are a directory stat() and lstat() are same */
    lxp_procpid_access,
    lxp_procpid_read,
    lxp_nodev_write,
    lxp_nodev_ioctl,
    lxp_nodev_fcntl
};

/*
 * Here we will simulate /proc/<pid> with the help of the real /proc/<pid>
 *
 * ## OpenServer problem here!!
 */

static int
lxp_procpid_open(int pid, int flags, int mode)
{
    char	filename[PATH_MAX];

    sprintf(filename, "/proc/%d", pid);
    PROCFS_TRACE_MSG(("TRACE: lxp_procpid_open(%s)\n", filename));
    return open(filename, flags, mode);
}

static int
lxp_procpid_stat(int pid, struct stat *buf)
{
    char	filename[PATH_MAX];

    sprintf(filename, "/proc/%d", pid);
    PROCFS_TRACE_MSG(("TRACE: lxp_procpid_stat(%s)\n", filename));
    return stat(filename, buf);
}

static int
lxp_procpid_access(int pid, int mode)
{
    char	filename[PATH_MAX];

    sprintf(filename, "/proc/%d", pid);
    PROCFS_TRACE_MSG(("TRACE: lxp_procpid_access(%s)\n", filename));
    return access(filename, mode);
}

static int
lxp_procpid_read(int pid, int fd, void *buf, int len)
{
    int		r;

    PROCFS_TRACE_MSG(("TRACE: lxp_procpid_read(%d)\n", fd));

    if ((r = read(fd, buf, len)) != -1)
	return r;

    /*
     * ## Need code here for simulating entries for all of our
     * pseudo-files like /proc/<pid>/cmdline
     */

    return 0;
}

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									 *
 *			     /proc/<pid>/cmdline			 *
 *									 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

/*
 * ## NOT YET
 * OpenServer cannot access the psargs without being root. 
 * Maybe we can pipe to `ps -ef` and parse for it. :-(
 */

#ifdef UW7

static int lxp_cmdline_open(int pid, int flags, int mode);
static int lxp_cmdline_stat(int pid, struct stat *buf);
static int lxp_cmdline_access(int pid, int mode);
static int lxp_cmdline_read(int pid, int fd, void *buf, int len);

static lxprocfs_t	lxp_cmdline_ops =
{
    "cmdline",
    lxp_cmdline_open,
    lxp_cmdline_stat,
    lxp_cmdline_stat,
    lxp_cmdline_access,
    lxp_cmdline_read,
    lxp_nodev_write,
    lxp_nodev_ioctl,
    lxp_nodev_fcntl
};

/*
 * Here we will simulate /proc/<pid>/cmdline with /proc/<pid>/psinfo
 *
 * ## OpenServer problem here!!
 */

static int
lxp_cmdline_open(int pid, int flags, int mode)
{
    char	filename[PATH_MAX];

    sprintf(filename, "/proc/%d/psinfo", pid);
    PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_open(/proc/%d/cmdline)\n", pid));
    return open(filename, O_RDONLY);
}

static int
lxp_cmdline_stat(int pid, struct stat *buf)
{
    struct stat		st;

    PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_stat(/proc/%d/cmdline)\n", pid));

    if (!buf)
    {
	PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_stat() NULL buf\n"));
	return enoany(EINVAL, "");
    }

    memset(&st, 0, sizeof(st));

    st.st_ino = 1328087050;
    st.st_mode = S_IFREG | S_IREAD | S_IRGRP | S_IROTH;
    st.st_nlink = 1;
    st.st_uid = getuid();
    st.st_gid = 0;
    st.st_rdev = 0;
    st.st_size = 0;
    st.st_blksize = 512;
    st.st_blocks = 0;

    st.st_atime =
    st.st_mtime =
    st.st_ctime = time(NULL);

    memcpy(buf, &st, sizeof(st));
    return 0;
}

static int
lxp_cmdline_access(int pid, int mode)
{
    PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_access(/proc/%d/cmdline) mode %d\n",
		      pid, mode));

    switch (mode)
    {
	case R_OK:
	case F_OK:
	    return 0;

	case X_OK:
	case W_OK:
	    return enoany(EPERM, "lxp_cmdline_access() {R,X}_OK - EPERM");

    }
    return enoany(EINVAL, "lxp_cmdline_access() unhandled mode");
}

static int
lxp_cmdline_read(int pid, int fd, void *buf, int len)
{
    struct psinfo	ps;
    int			r;
    off_t		addr;
    int			cnt;
    char		*tp;

    PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_read(/proc/%d/cmdline)\n", pid));

    addr = lseek(fd, 0, SEEK_CUR);	/* get the address into cmdline */
    lseek(fd, 0, SEEK_SET);		/* rewind /proc/<pid>/psinfo */

    if ((r = read(fd, &ps, sizeof(ps))) == -1)
    {
	lseek(fd, addr, SEEK_SET);
	PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_read() read fail: %s\n",
						    strerror(errno)));
	return -1;
    }

    lseek(fd, addr + r, SEEK_SET);	/* for next time */

    if (r != sizeof(ps))
    {
	PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_read() short read: %d\n", r));
	return enoany(EIO, "lxp_cmdline_read");
    }

    tp = ps.pr_psargs;
    for (cnt = 0; *tp && (cnt < sizeof(ps.pr_psargs)); ++tp)
    {
	++cnt;

	if (isspace(*tp))
	    *tp = '\0';
    }

    PROCFS_TRACE_MSG(("TRACE: lxp_cmdline_read() %d %d\n", addr, cnt));

    if (addr >= cnt)
	return 0;

    len = min(cnt-addr, len);
    memcpy(buf, &ps.pr_psargs[addr], len);
    return len;
}

#endif

/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									 *
 *			  Common Private Routines			 *
 *									 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 */

/*
 * The pathname parsing here needs to deal better strange paths
 * like:
 *	/proc/self/../self/.////./cmdline
 */

static lxprocfs_t *
lxprocfs_ops(const char *path, int *pidp)
{
    static lxprocfs_t	*lxp_ops[] =
    {
#ifdef UW7
	&lxp_cmdline_ops,
#endif
	NULL
    };
    lxprocfs_t	**op;
    int		pid;
    const char	*tp;

    /*
     * We own everything in /proc
     */

    if (strncmp(path, "/proc", 5) != 0)
	return NULL;

    if (!path[5])
	return &lxp_proc_ops;		/* The directory /proc */

    if (path[5] != '/')
	return NULL;			/* /procsomething/else */

    pid = 0;
    for (tp = &path[6]; isdigit(*tp); ++tp)
	pid = (pid * 10) + *tp - '0';
    
    if (!*tp)
    {
	if (pidp)
	    *pidp = pid;

	return &lxp_procpid_ops;	/* The directory /proc/<pid> */
    }

    if (*tp != '/')
	return NULL;			/* /proc49ers */

    /*
     * Something in the directory /proc/<pid>/
     */

    ++tp;
    for (op = lxp_ops; *op; ++op)
	if (strcmp(tp, (*op)->name) == 0)
	{
	    if (pidp)
		*pidp = pid;

	    return *op;
	}

    return NULL;	/* Not one of ours.  Maybe a native file */
}

static procfssw_t *
find_procfssw(int fd)
{
    procfssw_t	*fp;

    for (fp = procfs; fp; fp = fp->next)
	if (fp->fd == fd)
	    return fp;

    return NULL;
}

#endif /* OLD_PROCFS */
