#ifdef OSR5
#pragma comment(exestr, "%Z% %M% %I% %E% %Q%")
#endif

#ifdef SOLARIS
#pragma ident	"%Z%%M% %I% %D% SMI"
#endif
/*
 * Copyright (C) 1996 Michael Davidson.
 * All rights reserved.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#include	<assert.h>
#include	<errno.h>
#include	<signal.h>
#include	<stdlib.h>
#include	<string.h>
#include	<sys/wait.h>
#include	<sys/siginfo.h>
#include	<sys/regset.h>
#include	<sys/reg.h>
#include	<ucontext.h>

#include	"lxrun.h"
#include	"linux/resource.h"
#include	"linux/signal.h"
#include	"linux/sigcontext.h"
#include	"linux/wait.h"
#include	"trace.h"

extern void syscall_handler(int sig, siginfo_t *info, void *context);
extern int sendsig(void (*)(), int, void *);

int lx_signal(int lx_sig, lx_sighandler_t lx_handler);

#if defined(UNIXWARE) && !defined(UW7)
#define	sa_sigaction	sa_handler
#endif

/*
 * signals[] is a table of corresponding Linux and System V
 * error numbers - it is used by init_signals() to build the
 * lx_signo[] and os_signo[] lookup tables
 */
#   define SIG(name)	{ SIG##name, LX_SIG##name, "SIG" #name }

struct
{
	int	os_sig;
	int	lx_sig;
	const char	*name;
} signals[NSIG] =
{
	SIG(HUP),	SIG(INT),	SIG(QUIT),	SIG(ILL),
	SIG(TRAP),	SIG(ABRT),	SIG(BUS),	SIG(FPE),
	SIG(KILL),	SIG(USR1),	SIG(SEGV),	SIG(USR2),
	SIG(PIPE),	SIG(ALRM),	SIG(TERM),	SIG(CHLD),
	SIG(CONT),	SIG(STOP),	SIG(TSTP),	SIG(TTIN),
	SIG(TTOU),	SIG(XCPU),	SIG(XFSZ),	SIG(VTALRM),
	SIG(PROF),	SIG(WINCH),	SIG(PWR),
#if defined(UNIXWARE) || defined(SOLARIS)
							SIG(URG),
	/* gcc prefers the null initializers here to be in braces */
	SIG(IO),	{0, 0}, 	{0, 0},		{0, 0}
#else
							{0, 0},
	{0, 0},		{0, 0},		{0, 0},		{0, 0}
#endif
};

static int		lx_signo[NSIG+1];
static int		os_signo[NSIG+1];

typedef enum
{
	USE_SIGNAL = 0,
	USE_SIGACTION
} signal_use;

static struct
{
	struct sigaction	sig_action;
	void			(*sig_nal)();
	signal_use		use;
} signal_actions[NSIG + 1];

int
init_signals()
{
	int			i;
	struct sigaction	s;

	/*
	 * Initialise the signal number lookup tables
	 */

	SIG_TRACE_MSG(("TRACE: init_signals()\n"));

	for (i = 0; i < NSIG; i++)
	{
		int	os_sig	= signals[i].os_sig;
		int	lx_sig	= signals[i].lx_sig;

		assert(os_sig >= 0 && os_sig <= NSIG);
		assert(lx_sig >= 0 && lx_sig <= LX_NSIG);

		lx_signo[os_sig]	= lx_sig;
		os_signo[lx_sig]	= os_sig;
	}

	/*
	 * Install the system call handler for the int 0x80 instruction
	 * (which causes a general protection trap and results in SIGSEGV)
	 */

	sigemptyset(&s.sa_mask);
	s.sa_sigaction	= syscall_handler;
	s.sa_flags	= SA_NODEFER | SA_SIGINFO;

	signal_actions[SIGSEGV].use = USE_SIGACTION;

	return sigaction(SIGSEGV, &s, &signal_actions[SIGSEGV].sig_action);
}


static int
copy_in_sigmask(lx_sigset_t *from, sigset_t *to)
{
	int		i;
	unsigned long	lx_mask;

	if (! verify_read(from, sizeof(*from)))
		return -1;

	lx_mask = *from;
	sigemptyset(to);

	for (i = 1; lx_mask != 0; i++, lx_mask >>= 1)
		if ((lx_mask & 1)  && os_signo[i] != 0)
			sigaddset(to, os_signo[i]);

	return 0;
}


static int
copy_out_sigmask(sigset_t *from, lx_sigset_t *to)
{
	int		i;
	unsigned long	lx_mask = 0; /* otherwise appears uninitialized */

	if (! verify_write(to, sizeof(*to)))
		return -1;

	for (i = 0; (i < NSIG) && (signals[i].os_sig != 0); i++)
		if (sigismember(from, signals[i].os_sig))
			lx_mask |=  1 << (signals[i].lx_sig-1);

	*to = lx_mask;

	return 0;
}

void
deliver_signal(
	int		sig,
	siginfo_t	*info,
	void		*context
	)
{
	int			save		= errno;
	struct sigaction	*act;
	void			(*hand)();
	signal_use		use;
	unsigned long		*regs;
	struct lx_sigcontext_struct	lx_context;

#if defined(OSR5)
	regs	= (unsigned long *)((ucontext_t *)context)->uc_mcontext.regs;
#else
	regs	= (unsigned long *)((ucontext_t *)context)->uc_mcontext.gregs;
#endif

        use  	= signal_actions[sig].use;

	TRACE_CALL( TRACE_SIG, trace_signal(sig, info, context));
	
	if (use == USE_SIGACTION)
	{
		act = &signal_actions[sig].sig_action;

		if ((act->sa_handler == SIG_DFL) ||
		    (act->sa_handler == SIG_IGN))
		{
		sigaction(sig, act, NULL);
			return;			
	}
	}
	else
	{
		hand = signal_actions[sig].sig_nal;

		if ((hand == SIG_DFL) || (hand == SIG_IGN))
		{
			signal(sig, hand);
			return;
		}
	}

		/*
		 * set up Linux signal handler stack frame
		 */
		lx_context.gs		= regs[GS];
		lx_context.fs		= regs[FS];
		lx_context.es		= regs[ES];
		lx_context.ds		= regs[DS];
		lx_context.edi		= regs[EDI];
		lx_context.esi		= regs[ESI];
		lx_context.ebp		= regs[EBP];
		lx_context.esp		= regs[ESP];
		lx_context.ebx		= regs[EBX];
		lx_context.edx		= regs[EDX];
		lx_context.ecx		= regs[ECX];
		lx_context.eax		= regs[EAX];
		lx_context.trapno	= 0;
		lx_context.err		= 0;
		lx_context.eip		= regs[EIP];
		lx_context.cs		= regs[CS];
		lx_context.eflags	= regs[EFL];
		lx_context.esp_at_signal= regs[ESP];
		lx_context.ss		= regs[SS];
		lx_context.i387		= 0;
		lx_context.oldmask	= 0;
		lx_context.cr2	= 0;

	if (use == USE_SIGACTION)
	{
	        SIG_TRACE_MSG((
	        "TRACE: deliver sigaction: sendsig(%#x, %d -> %d = %s, %#x)\n",
				act->sa_handler,
				sig, lx_signo[sig], os_sigName(sig),
				&lx_context)); 

		sendsig(act->sa_handler, lx_signo[sig], &lx_context);
	}
	else
	{
		SIG_TRACE_MSG((
		    "TRACE: deliver signal: sendsig(%#x, %d %d = %s, %#x)\n",
				hand,
				sig, lx_signo[sig], os_sigName(sig),
				&lx_context)); 

		sendsig(hand, lx_signo[sig], &lx_context);	
	}
	
	/*
	 * copy back any changes
	 */
	TRACE_CALL( TRACE_SIG,	trace_reg_changes(regs, &lx_context));

	regs[EDI]	= lx_context.edi;
	regs[ESI]	= lx_context.esi;
	regs[EBP]	= lx_context.ebp;
	regs[ESP]	= lx_context.esp;
	regs[EBX]	= lx_context.ebx;
	regs[EDX]	= lx_context.edx;
	regs[ECX]	= lx_context.ecx;
	regs[EAX]	= lx_context.eax;
	regs[EIP]	= lx_context.eip;
	regs[EFL]	= lx_context.eflags;

	errno = save;
}


int
lx_sigaction(int lx_sig, struct lx_sigaction *lx_new_act,
					struct lx_sigaction *lx_old_act)
{
	int			sig;
	struct sigaction	new_action;
	struct sigaction	old_action;
	struct sigaction	*new_act	= NULL;
	struct sigaction	*old_act	= NULL;

	SIG_TRACE_MSG(("TRACE: sigaction(%s)\n", lx_sigName(lx_sig)));

	memset(&new_action, 0, sizeof(new_action));
	memset(&old_action, 0, sizeof(old_action));

	if (lx_sig <= 0 || lx_sig > LX_NSIG)
	{
	    return enoany(EINVAL, "sigaction() - sig <=0 or >LX_NSIG");
	}

	sig = os_signo[lx_sig];

	if (lx_new_act)
	{
		/*
		 * linux implements signal() on top of sigaction, using
		 * the one-shot flag.
		 */
		if (lx_new_act->flags & LX_SA_ONESHOT)
			return lx_signal(lx_sig, lx_new_act->handler);

		if (! verify_read(lx_new_act, sizeof(*lx_new_act)))
			return -1;

		new_act	= &new_action;

		if (lx_new_act->handler == LX_SIG_DFL)
			new_act->sa_handler = SIG_DFL;
		else if (lx_new_act->handler == LX_SIG_IGN)
			new_act->sa_handler = SIG_IGN;
		else
			new_act->sa_sigaction = deliver_signal;

		copy_in_sigmask(&lx_new_act->mask, &new_act->sa_mask);
		/*
		 * make sure SIGSEGV is never blocked
		 */
		sigdelset(&new_act->sa_mask, SIGSEGV);

		/*
		 * WARNING WARNING WARNING!!!!!!
		 *
		 * The SA_NODEFER flag appears to enable multiple signals
		 * to work (bash, tcsh, bsh, and zsh work). Otherwise, only
		 * one signal appears to be presented to the application,
		 * after which the signal acts like it is SIG_IGN'ed.
		 *
		 * FIX ME OR FIGURE ME OUT!!!!!
		 */
		new_act->sa_flags	= SA_SIGINFO | SA_NODEFER;

		if (lx_new_act->flags & LX_SA_NOCLDSTOP)
			new_act->sa_flags	|= SA_NOCLDSTOP;
		if (lx_new_act->flags & LX_SA_STACK)
			new_act->sa_flags	|= SA_ONSTACK;
		if (lx_new_act->flags & LX_SA_RESTART)
			new_act->sa_flags	|= SA_RESTART;
		if (lx_new_act->flags & LX_SA_ONESHOT)
			new_act->sa_flags	|= SA_RESETHAND;
		if (lx_new_act->flags & LX_SA_NOMASK)
			new_act->sa_flags	|= SA_NODEFER;
	}

	if (lx_old_act)
		old_act	= &old_action;

	/*
	 * if signal is SIGSEGV just retrieve the old action
	 * from the emulator's copy, but don't actually call
	 * sigaction()
	 */
	if (sig == SIGSEGV)
	{
		if (old_act)
			*old_act = signal_actions[sig].sig_action;
	}
	else
	{
		if (sigaction(sig, new_act, old_act) != 0)
			return -1;
	}

	if (old_act)
	{
		if (! verify_write(lx_old_act, sizeof(*lx_old_act)))
			return -1;

		if (old_act->sa_handler == SIG_DFL)
			lx_old_act->handler = LX_SIG_DFL;
		else if (old_act->sa_handler == SIG_IGN)
			lx_old_act->handler = LX_SIG_IGN;
		else
			lx_old_act->handler =
				signal_actions[sig].sig_action.sa_handler;

		copy_out_sigmask(&old_act->sa_mask, &lx_old_act->mask);

		lx_old_act->flags	= 0;
		if (old_act->sa_flags & SA_NOCLDSTOP)
			lx_old_act->flags	|= LX_SA_NOCLDSTOP;
		if (old_act->sa_flags & SA_ONSTACK)
			lx_old_act->flags	|= LX_SA_STACK;
		if (old_act->sa_flags & SA_RESTART)
			lx_old_act->flags	|= LX_SA_RESTART;
		if (old_act->sa_flags & SA_NODEFER)
			lx_old_act->flags	|= LX_SA_NOMASK;
		if (old_act->sa_flags & SA_RESETHAND)
			lx_old_act->flags	|= LX_SA_ONESHOT;
	}

	/*
	 * finally update the current action for this signal
	 */
	if (new_act)
	{
		if (new_act->sa_sigaction == deliver_signal)
			new_act->sa_handler = lx_new_act->handler;

		signal_actions[sig].sig_action = *new_act;

		signal_actions[sig].use = USE_SIGACTION;
	}

	return 0;
}	


int
lx_signal(int lx_sig, lx_sighandler_t lx_handler)
{
	lx_sighandler_t	odisp;
	int		sig;

	SIG_TRACE_MSG(("TRACE: lx_signal(%s, 0x%x)\n",
			lx_sigName(lx_sig), lx_handler));

	if (lx_sig <= 0 || lx_sig > LX_NSIG)
	{
	    return enoany(EINVAL, "lx_signal() - sig <=0 or >LX_NSIG");
	}

	sig = os_signo[lx_sig];

	switch ((unsigned int)lx_handler)
	{
	    case (unsigned int)LX_SIG_DFL:
		lx_handler = SIG_DFL;
		break;

	    case (unsigned int)LX_SIG_IGN:
		lx_handler = SIG_IGN;
		break;

	    default:
		if (!verify_read((void *)lx_handler, sizeof(lx_handler)))
	return -1;
	}

	if (sig != SIGSEGV)
		if ((odisp = signal(sig, lx_handler)) == (lx_sighandler_t)-1)
			return -1;

	/*
	 * finally update the current action for this signal
	 */
	signal_actions[sig].sig_nal = lx_handler;
	signal_actions[sig].use = USE_SIGNAL;

	return (int)odisp;
}	


int lx_sigreturn()	{ return enosys("sigreturn"); }

/*
 * Linux sigsuspend() is weird ....
 */
int
lx_sigsuspend(int restart, unsigned long lx_oldmask, unsigned long lx_set)
{
	sigset_t	mask;

	copy_in_sigmask(&lx_set, &mask);
	/*
	 * make sure SIGSEGV is never blocked
	 */
	sigdelset(&mask, SIGSEGV);

	return sigsuspend(&mask);
}	


int
lx_sigpending(lx_sigset_t *lx_mask)
{
	sigset_t	mask;
	int		r;

	if ((r = sigpending(&mask)) == 0)
		copy_out_sigmask(&mask, lx_mask);

	return r;
}	


int
lx_sigprocmask(int lx_cmd, lx_sigset_t *lx_new_mask, lx_sigset_t *lx_old_mask)
{
	int		cmd		= 0;
	sigset_t	omask;
	sigset_t	nmask;
	sigset_t	*old_mask	= NULL;
	sigset_t	*new_mask	= NULL;

	if (lx_new_mask)
	{
		switch (lx_cmd)
		{
			case LX_SIG_BLOCK:
				cmd = SIG_BLOCK;
				break;
			case LX_SIG_UNBLOCK:
				cmd = SIG_UNBLOCK;
				break;
			case LX_SIG_SETMASK:
				cmd = SIG_SETMASK;
				break;
			default:
			    return enoany(EINVAL,
					  "lx_sigprocmask() unknown cmd");
		}

		new_mask = &nmask;

		if (copy_in_sigmask(lx_new_mask, new_mask) != 0)
			return -1;
		/*
		 * make sure SIGSEGV is never blocked
		 */
		sigdelset(new_mask, SIGSEGV);
	}

	if (lx_old_mask)
		old_mask = &omask;

	if (sigprocmask(cmd, new_mask, old_mask) != 0)
		return -1;

	if (lx_old_mask && (copy_out_sigmask(old_mask, lx_old_mask) != 0))
		return -1;

	return 0;
}


int
lx_sgetmask()
{
	lx_sigset_t	omask;

	lx_sigprocmask(LX_SIG_SETMASK, NULL, &omask);

	return omask;
}


int
lx_ssetmask(lx_sigset_t nmask)
{
	lx_sigset_t	omask;

	lx_sigprocmask(LX_SIG_SETMASK, &nmask, &omask);

	return omask;
}


int
lx_kill(int pid, int lx_sig)
{
	return kill(pid, os_signo[lx_sig]);
}


int
lx_wait4(int pid, int *lx_stat_loc, int lx_options, struct lx_rusage *lx_ru)
{
	int		options	= 0;
	int		r;
	int		status;

	if (lx_options & LX_WNOHANG)
		options |= WNOHANG;
	if (lx_options & LX_WUNTRACED)
		options |= WUNTRACED;

	r = waitpid(pid, &status, options);

	/*
	 * convert the status to Linux format
	 * the only real difference here is the signal #
	 */
	if (r != -1 && lx_stat_loc)
	{
		int	lx_status = 0;

		if (WIFEXITED(status))
			lx_status = WEXITSTATUS(status) << 8;
		else if (WIFSIGNALED(status))
		{
			lx_status = lx_signo[WTERMSIG(status)];
			if (WCOREDUMP(status))
				lx_status |= 0x80;
		}
		else if (WIFSTOPPED(status))
			lx_status = (lx_signo[WSTOPSIG(status)] << 8) | 0x7f;

		if (! verify_write(lx_stat_loc, sizeof(*lx_stat_loc)))
			return -1;

		*lx_stat_loc = lx_status;
	}

	if (lx_ru)	/* resource usage NOT IMPLEMENTED */
	{
		if (! verify_write(lx_ru, sizeof(*lx_ru)))
			return -1;

		memset(lx_ru, 0, sizeof(*lx_ru));
	}

	return r;
}	


int
lx_waitpid(int pid, int *lx_stat_loc, int lx_options)
{ 
	return lx_wait4(pid, lx_stat_loc, lx_options, NULL);
}


#include	<stdio.h>

const char *
lx_sigName(int sig)
{
    int		n;
    static char	buf[32];

    for (n = 0; (n < NSIG) && (signals[n].os_sig != 0); ++n)
	if (signals[n].lx_sig == sig)
	    return signals[n].name;

    sprintf(buf, "<Unknown %d>", sig);
    return buf;
}

const char *
os_sigName(int sig)
{
    int		n;
    static char	buf[32];

    for (n = 0; (n < NSIG) && (signals[n].os_sig != 0); ++n)
	if (signals[n].os_sig == sig)
	    return signals[n].name;

    sprintf(buf, "<Unknown %d>", sig);
    return buf;
}

